/* $Id: buffer.c,v 1.5 2012/12/13 10:35:57 wg Exp $
 * Copyright Wolfram Gloger <evn-tool@malloc.de> Jul 2011, Sep 2011, Aug 2012
 * This program is published under the GNU General Public license
 */

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <assert.h>
#include <unistd.h>

#include "mpegts.h"

#define BUFFER_PACKETS 2048
#define VERBOSITY      0

/*
 * Do linear pcr interpolation from buffer index i0 through i1, inclusive.
 * 0 <= i0 <= i1 < 2*BUFFER_PACKETS
 */
static void
lin_interpol(int64_t pcr0, int pix0, int64_t pcr1, int pix1,
	     int i0, int i1, int64_t out[2*BUFFER_PACKETS])
{
    int64_t pcrs = pcr0*(pix1-pix0) + (i0-pix0)*(pcr1-pcr0);

    assert(0 <= i0);
    //assert(i0 <= i1);
    assert(i1 < 2*BUFFER_PACKETS);

    if (pcr0 > pcr1) {
	// wrap: linear interpolation impossible
	while (i0 < i1)
	    out[i0++] = pcr0++;
	out[i1] = pcr1;
	return;
    }

    while (i0 <= i1) {
	int64_t pcr = pcrs/(pix1-pix0);
#if 0
	fprintf(stderr, "Pkt %5d -> ", i0);
	print_pcr(stderr, pcr);
	fprintf(stderr, "\n");
#endif
	out[i0] = pcr;
	pcrs += pcr1-pcr0;
	++i0;
    }
}

// 4 extra bytes at the start for timestamp insertion
static uint8_t rbuffer[4 + 2*TS_PACKET_SIZE*BUFFER_PACKETS];

#define GET_BUFFER(buf_i, i) \
(rbuffer + 4 + (((buf_i)*BUFFER_PACKETS+(i))%(2*BUFFER_PACKETS))*TS_PACKET_SIZE)

long
buffer_process_packets(struct TSContext* ts,
		       int (*output_cb)(void* arg,
					uint8_t* pkt, int64_t pcr, int fl, long pkt_n),
		       void* arg0)
{
    long total_np, out_np;
    int i, np, buf_i = 0;
    int     pix0 = -1, prev_pix0;
    int64_t pcr0 = -1, prev_pcr0 = -1;

    int64_t pcri[2*BUFFER_PACKETS] = { -1 };
    int     keyf[2*BUFFER_PACKETS] = { 0 };

    int parse_i = 0; // the index from which parsing needs to start
    uint8_t *pkt;
    int64_t off0;

    // read initial packets
    np = mpegts_read_packets(ts, GET_BUFFER(buf_i, 0), BUFFER_PACKETS);
    off0 = mpegts_get_packet_offset(ts);

    for (out_np = total_np = 0;; total_np += BUFFER_PACKETS) {
	int do_extrapolate = 0;

	// The buffer buf_i contains BUFFER_PACKETS consecutive packets,
	// with the last pcr value being pcr0 at index pix0,
	// pix0 < BUFFER_PACKETS

	// read next block of packets
	int np2 = mpegts_read_packets(ts, GET_BUFFER(1-buf_i, 0), BUFFER_PACKETS);
	if (np2 > 0)
	    np += np2;
	//fprintf(stderr, "==== Got %d packs\n", np);

	// find next pcr's and interpolate as far as possible
	for (i = pix0+1; i < np; ++i) {
	    int64_t pcr = get_pcr(GET_BUFFER(buf_i, i));
	    if (pcr >= 0) {
		prev_pcr0 = pcr0;
		prev_pix0 = pix0;
		pcr0 = pcr;
		pix0 = i;
		if (VERBOSITY>0 && total_np == 0) {
		    fprintf(stderr, "Packet %7ld: PCR=", total_np+i);
		    print_pcr(stderr, pcr);
		    fprintf(stderr, " prev=");
		    print_pcr(stderr, prev_pcr0);
		    fprintf(stderr, " prev_pix0=%d\n", prev_pix0);
		}

		if (prev_pcr0 >= 0)
		    lin_interpol(prev_pcr0, prev_pix0, pcr, i,
				 (pcri[0]<0 ? 0 : prev_pix0+1), i, pcri);
	    }
	}

	if (np < 2*BUFFER_PACKETS) {
	    if (VERBOSITY > 0)
		fprintf(stderr, "** EOF reached,");
	    do_extrapolate = 1;
	} else if (pix0 < BUFFER_PACKETS) {
	    fprintf(stderr, "** pcr not found in buffer,");
	    do_extrapolate = 2;
	}

	if (do_extrapolate) {
	    if (prev_pcr0 >= 0) {
		if (VERBOSITY > 0)
		    fprintf(stderr, " extrapolating from %d-%d.\n", pix0+1, np-1);
		assert(prev_pix0 < pix0);
		lin_interpol(prev_pcr0, prev_pix0, pcr0, pix0, pix0+1, np-1, pcri);
		pix0 = np-1;
		pcr0 = pcri[pix0];
	    } else {
		if (pix0 == -1) {
		    fprintf(stderr, " no initial pcr found, using 0, 1, 2...\n");
		    pcr0 = 0;
		} else
		    fprintf(stderr, " using 1 for increment.\n");
		while (pix0 < np-1)
		    pcri[++pix0] = ++pcr0;
	    }
#if 0
	    for (i=0; i<np; ++i) {
		fprintf(stderr, "Packet %7zd: PCR=", total_np+i);
		print_pcr(stderr, pcri[i]);
		fprintf(stderr, "\n");
	    }
#endif
	}

	// We now have interpolated pcr's from parse_i through pix0.
	// Parse those packets.
	while (parse_i <= pix0) {
	    pkt = GET_BUFFER(buf_i, parse_i);
	    if (parse_i > 0) {
		// check for pcr wraparound
		//assert(pcri[parse_i] > pcri[parse_i-1]);
		if (0 && pcri[parse_i] <= pcri[parse_i-1])
		    return total_np + np;
	    }
	    if (mpegts_parse_packet(ts, pkt, pcri[parse_i])) {
		int64_t ko = mpegts_get_keyframe_offset(ts);

		if (ko < off0)
		    fprintf(stderr, "! keyframe before off0\n");
		else {
		    ko -= off0;
		    assert(ko % TS_PACKET_SIZE == 0);
		    ko /= TS_PACKET_SIZE;
		    if (0)
			fprintf(stderr, "key: %9"PRId64" buf0: %9"PRId64"\n",
				ko, (int64_t)total_np);
		    if (ko < total_np) {
			fprintf(stderr, "! keyframe before total_np=%ld\n", total_np);
		    } else if (ko - total_np < 2*BUFFER_PACKETS) {
			//fprintf(stderr, "marking buf %d as keyframe\n", (int)(ko-total_np));
			keyf[ko - total_np] = 1;
		    } else
			fprintf(stderr, "! can't mark buf %"PRId64" as keyframe\n",
				ko-total_np);
		}
	    }
	    ++parse_i;
	}

	if (np < 2*BUFFER_PACKETS)
	    break;

	// output packets with interpolated pcr
	pkt = GET_BUFFER(buf_i, 0);
	for (i = 0; i < BUFFER_PACKETS; ++i) {
	    int out_res = output_cb(arg0, pkt, pcri[i], keyf[i], total_np+i);
	    if (out_res < 0) {
		if (VERBOSITY > 0)
		    fprintf(stderr, "End output offset: %9"PRId64"\n",
			    mpegts_get_offset(ts) - (2*BUFFER_PACKETS - i)*TS_PACKET_SIZE);
		return total_np + np;
	    }
	    out_np += out_res;
	    if (out_np == 1) {
		fprintf(stderr, "Begin output offset: %9"PRId64"\n",
			mpegts_get_offset(ts) - (2*BUFFER_PACKETS - i)*TS_PACKET_SIZE);
	    }
	    pkt += TS_PACKET_SIZE;
	}

	// slide window / flip buffer
	assert(parse_i >= BUFFER_PACKETS);
	parse_i   -= BUFFER_PACKETS;
	prev_pix0 -= BUFFER_PACKETS;
	pix0      -= BUFFER_PACKETS;
	np        -= BUFFER_PACKETS;
	for (i=0; i <= pix0; ++i) {
	    pcri[i] = pcri[BUFFER_PACKETS+i];
	    keyf[i] = keyf[BUFFER_PACKETS+i];
	}
	while (i < 2*BUFFER_PACKETS) {
	    pcri[i] = -1;
	    keyf[i] = 0;
	    ++i;
	}

	buf_i = 1-buf_i;
    }

    // output remaining packets
    for (i=0; i<np; ++i) {
	pkt = GET_BUFFER(buf_i, i);
	if (output_cb(arg0, pkt, pcri[i], keyf[i], total_np+i) < 0)
	    break;
    }

    return total_np+np;
}
