/*
 * $Id: main.c,v 1.24 2013/04/05 17:59:39 wg Exp $
 * Copyright 2011-2013 by Wolfram Gloger
 *
 *  This program is published under the GNU General Public license
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <assert.h>
#include <limits.h>

#include <unistd.h>
#include <fcntl.h>
#include <utime.h>

#include <netinet/in.h>

#include "mpegts.h"
#include "buffer.h"
#include "mpegts-process.h"
#include "endian.h"

static void
print_pts(FILE* f, int64_t pts)
{
    int64_t secE4 = pts/9;

    //fprintf(f, "%8"PRIx64" ", pts);
    if (pts < 0)
	fprintf(f, "---INVALID---");
    else
	fprintf(f, "%8"PRId64".%04"PRId64, secE4/10000, secE4%10000);
}

static int64_t
pts_to_pcr(int64_t pts)
{
    return pts*300;
}

/* The arrival_time_stamp consists of 4 bytes. The first two bits are
   reserved for a copyright and should be set to 11 (binary) */ 
static void
write_starttime_code(unsigned char* q, int64_t pcr)
{
    *q = 3 << 6;
    *q++ |= (pcr >> 24) & 0xFF; 
    *q++ =  (pcr >> 16) & 0xFF;
    *q++ =  (pcr >> 8)  & 0xFF;
    *q++ =  pcr;
}

// callback for writing packets
struct output_packet_s {
    int64_t pcr0;
    int64_t pcr1;
    int64_t pcr_first;
    int64_t pcr_last;
    int fd;
    int insert_timestamp;
    long out_np;
    int do_log;
    int wait_keyframe; // set to 0 after keyframe found
    long n_keyframe;
    int wait_smaller_pcr1;
};

// output is terminated when <0 is returned
static int
output_packet(void* arg0,
	      unsigned char* pkt, int64_t pcr, int fl, long pkt_n)
{
    struct output_packet_s* ops = arg0;
    size_t wr_size;

    // wait for start
    if (ops->out_np == 0 && pcr < ops->pcr0)
	return 0;

    // check for end
    if (ops->pcr1 > 0) {
	if (!ops->wait_smaller_pcr1) {
#if 0
	    print_pcr(stderr, pcr);
	    fprintf(stderr, " pcr1=");
	    print_pcr(stderr, ops->pcr1);
	    fprintf(stderr, " %ld\n", pkt_n);
#endif
	    if (ops->pcr1 > 0 && pcr >= ops->pcr1)
		return -1;
	} else {
	    if (pcr < ops->pcr1) {
		ops->wait_smaller_pcr1 = 0;
		fprintf(stderr, "wait_smaller: pcr=");
		print_pcr(stderr, pcr);
		fprintf(stderr, "\n");
	    }
	}
    }

    if (ops->wait_keyframe) {
	if (fl & 1) {
	    ops->wait_keyframe = 0;
	    fprintf(stderr, "waited for keyframe until pkt %ld\n", pkt_n);
	} else {
	    //fprintf(stderr, "waiting for keyframe at %ld\n", pkt_n);
	    return 0;
	}
    }

    // statistics
    if (ops->out_np == 0)
	ops->pcr_first = pcr;
    ops->pcr_last = pcr;
    if (fl & 1)
	++ops->n_keyframe;

    if (ops->insert_timestamp) {
	pkt -= 4;
	write_starttime_code(pkt, pcr);
	wr_size = TS_PACKET_SIZE+4;
    } else
	wr_size = TS_PACKET_SIZE;

    if (ops->do_log) {
	fprintf(stderr, "Out%7ld: ", ops->out_np);
	print_pcr(stderr, pcr);
	fprintf(stderr, "\n");
    }
    if (write(ops->fd, pkt, wr_size) < wr_size) {
	fprintf(stderr, "Can't write output.\n");
	return -1;
    }
    ++ops->out_np;
    return 1;
}

/***********************************************************************/

int
mpegts_cut(struct TSContext *ts0, struct output_packet_s* ops, const char* f,
	   int wait_keyframe)
{
    long total_np = 0;
    int seek_res;

    ops->pcr_first = ops->pcr_last = 0;
    ops->out_np    = ops->n_keyframe = 0;
    ops->wait_keyframe = wait_keyframe;
    if (f && !strcmp(f, "-")) {
	ops->fd = 1;
    } else if (f && *f) {
	ops->fd = open(f, O_WRONLY|O_CREAT|O_TRUNC, 0644);
	if (ops->fd < 0) {
	    fprintf(stderr, "Can't open output %s\n", f);
	    return -1;
	}
    }

    if (ops->pcr1 > 0 && ops->pcr1 < ops->pcr0) {
	fprintf(stderr, "pcr1 < pcr0, handle wrap\n");
	ops->wait_smaller_pcr1 = 1;
    }

    // seek to start
    if (ops->pcr0 != -1) {
	seek_res = mpegts_seek_pcr(ts0, ops->pcr0);
	if (seek_res) {
	    fprintf(stderr, "seeking failed seek_res=%d\n", seek_res);
	}
    }

    // copy packets to output
    if (ops->fd > 0)
	total_np += buffer_process_packets(ts0, output_packet, ops);

    if (ops->fd > 1) {
	struct utimbuf ut;

	close(ops->fd);
	if (mpegts_get_utime(ts0, &ut) == 0)
	    utime(f, &ut);
    }

    fprintf(stderr,
	    "------------------------\n"
	    "Read     :%8ld\n"
	    "Written  :%8ld\n"
	    "Keyframes:%8ld\n",
	    total_np, ops->out_np, ops->n_keyframe);
    fprintf(stderr,       "First PCR: "); print_pcr(stderr, ops->pcr_first);
    fprintf(stderr,     "\nLast PCR : "); print_pcr(stderr, ops->pcr_last);
    if (ops->n_keyframe > 0) {
	fprintf(stderr, "\nSec/Keyfr: ");
	print_pcr(stderr, (ops->pcr_last-ops->pcr_first)/ops->n_keyframe);
    }
    fprintf(stderr, "\n");

    return 0;
}

int mpegts_cut_dvbapp(struct TSContext *ts0, struct output_packet_s* ops,
		      const char* f, int wait_keyframe,
		      int64_t first_pcr, int64_t last_pcr,
		      char* out_file, char* out_s)
{
    char in_file[PATH_MAX], *in_s;
    int fd, i, i_max = -1, out_n = 0;
    struct {
	uint64_t where;
	uint32_t what;
    } cuts_list[128];
    char cuts_buf[sizeof(uint64_t)];
    int cuts_n = 0;
    uint64_t delta_max;

    strncpy(in_file, f, PATH_MAX-6);
    in_file[PATH_MAX-6] = '\0';
    in_s = in_file + strlen(in_file);

    // read dvbapp '.cuts' file
    strcpy(in_s, ".cuts");
    fd = open(in_file, O_RDONLY);
    if (fd < 0) {
	fprintf(stderr, "Can't open cuts %s\n", in_file);
	return -1;
    }

    while (cuts_n < 128 &&
	   read(fd, cuts_buf, sizeof cuts_buf) == sizeof cuts_buf &&
	   read(fd, &cuts_list[cuts_n].what, sizeof(uint32_t)) == sizeof(uint32_t)) {
	cuts_list[cuts_n].where = AV_RB64(cuts_buf);
	cuts_list[cuts_n].what  = ntohl(cuts_list[cuts_n].what);

	fprintf(stderr, "%3d ", cuts_n);
	print_pts(stderr, cuts_list[cuts_n].where);
	fprintf(stderr, " %ld\n", (long)cuts_list[cuts_n].what);

	if (cuts_list[cuts_n].what < 3)
	    ++cuts_n;
    }
    close(fd);

    delta_max = 0;
    for (i=0; i<cuts_n-1; ++i) {
	uint64_t d = cuts_list[i+1].where - cuts_list[i].where;

	if (d > delta_max) {
	    delta_max = d;
	    i_max = i;
	}
    }
    // check final section too
    if (cuts_n > 0) {
	int64_t dpcr = last_pcr - (first_pcr + pts_to_pcr(cuts_list[cuts_n-1].where));

	if (dpcr > 0 && pts_to_pcr(delta_max) < dpcr)
	    i_max = cuts_n - 1;
    }

    if (i_max >= 0) {
	ops->pcr0 =
	    first_pcr + pts_to_pcr(cuts_list[i_max  ].where);
	ops->pcr1 = i_max<(cuts_n-1) ?
	    first_pcr + pts_to_pcr(cuts_list[i_max+1].where) : -1;
	fprintf(stderr, "Cuts list in: ");
	print_pcr(stderr, ops->pcr0);
	fprintf(stderr, " out: ");
	print_pcr(stderr, ops->pcr1);
	fprintf(stderr, "\n");

	sprintf(out_s, "-%02d.ts", out_n++);
	if (!access(out_file, F_OK)) {
	    fprintf(stderr, "* Skipping existing file %s\n", out_file);
	} else {
	    if (mpegts_cut(ts0, ops, out_file, wait_keyframe) < 0)
		return -2;
	}
	
    }

    return 0;
}

int
main(int argc, char** argv)
{
    struct output_packet_s ops = {
	.pcr0 = -1,
	.pcr1 = -1,
	.fd = -1,
    };
    int pcr_relative = 0;

    int fix_pat = 1;
    // The probe size for finding PAT and PMT.
    int pat_probe_size = 5000000;
    struct count_packet_s cps;
    int cue_sheet = 0;
    int wait_keyframe = 1;
    struct TSContext* ts0;

    while (argc > 1 && argv[1][0] == '-' && argv[1][1]) {
	if (!strncmp(argv[1], "-r", 2))
	    pcr_relative = 1;
	if (!strncmp(argv[1], "-ss", 3)) {
	    ops.pcr0 = atopcr(&argv[1][3]);
	}
	if (!strncmp(argv[1], "-tt", 3)) {
	    ops.pcr1 = atopcr(&argv[1][3]);
	}
	if (!strncmp(argv[1], "-nofixpat", 9))
	    fix_pat = 0;
	if (!strncmp(argv[1], "-ppsize", 7))
	    pat_probe_size = atoi(&argv[1][7]);
	if (!strncmp(argv[1], "-m2ts", 5))
	    ops.insert_timestamp = 1;
	if (!strncmp(argv[1], "-d", 2))
	    ops.do_log = 1;
	if (!strncmp(argv[1], "-nokeyframe", 11))
	    wait_keyframe = 0;
	if (!strncmp(argv[1], "-cuts", 5))
	    cue_sheet = 2;
	else if (!strncmp(argv[1], "-c", 2))
	    cue_sheet = 1;
	++argv;
	--argc;
    }

    if (argc < 2) {
	fprintf(stderr,
		"Usage: ts-cut [-r] [-ss<pcr0>] [-tt<pcr1>] [-c] [-ppsize<probe_size>]\n"
		"         input.ts output.ts\n");
	exit(1);
    }

    if (pcr_relative && pat_probe_size < 2000000)
	pat_probe_size = 2000000;

    ts0 = mpegts_new_context(argv[1]);
    if (!ts0)
	exit(2);

    if (pat_probe_size > 0) {
	long np_read = mpegts_get_info(ts0, pat_probe_size, &cps), np_histo;

	if (np_read < 0)
	    fprintf(stderr, "Can't get mpegts info (%ld)\n", np_read);
	else {
	    int pid;

	    fprintf(stderr, "Probed %ld packets.\n", np_read);
	    fprintf(stderr, "First PCR: ");
	    print_pcr(stderr, cps.first_pcr);
	    fprintf(stderr, "\nLast PCR:  ");
	    print_pcr(stderr, cps.last_pcr);

	    fprintf(stderr, "\nPID-----Histo\n");
	    for (np_histo=0, pid=0; pid<8192; ++pid) {
		if (cps.pid_histo[pid]) {
		    np_histo += cps.pid_histo[pid];
		    fprintf(stderr, "%4x %8ld\n", pid, cps.pid_histo[pid]);
		}
	    }
	    fprintf(stderr, "-------------\n");
	    fprintf(stderr, "Sum   %7ld (%ld keyframes)\n", np_histo, cps.n_keyframe);
	    if (fix_pat)
		mpegts_compute_best_sid(ts0, cps.pid_histo);

	    mpegts_print_pmt(ts0);

	    if (cps.n_keyframe == 0 && wait_keyframe) {
		fprintf(stderr, "Waiting for keyframe disabled because no keyframes found\n");
		wait_keyframe = 0;
	    }
	}
    }

    if (argv[2]) {
	char out_file[PATH_MAX], *out_s;

	strncpy(out_file, argv[2], PATH_MAX-12);
	out_file[PATH_MAX-12] = '\0';
	out_s = out_file + strlen(out_file);
	if ((out_s - out_file) > 3 && !memcmp(out_s-3, ".ts", 3))
	    out_s -= 3;

	if (pcr_relative) {
	    if (ops.pcr0 != -1)
		ops.pcr0 += cps.first_pcr;
	    if (ops.pcr1 != -1)
		ops.pcr1 += cps.first_pcr;
	}

	switch (cue_sheet) {
	case 0:
	    mpegts_cut(ts0, &ops, argv[2], wait_keyframe);
	    break;
	case 1: {
	    int out_n = 0, wait_for_start = (ops.pcr0 == -1);
	    char cline[256];

	    while (fgets(cline, sizeof cline, stdin)) {
		int skip = 0;
		int64_t cpcr;

		switch (cline[0]) {
		case '#': case '\n':
		    continue;
		case '!':
		    skip = 1;
		}

		cpcr = atopcr(cline+skip);
		if (pcr_relative)
		    cpcr += cps.first_pcr;

		if (wait_for_start) {
		    ops.pcr0 = cpcr;
		    wait_for_start = 0;
		    continue;
		}

		ops.pcr1 = cpcr;
		sprintf(out_s, "-%02d.ts", out_n++);
		if (!access(out_file, F_OK)) {
		    fprintf(stderr, "* Skipping existing file %s\n", out_file);
		} else {
		    if (mpegts_cut(ts0, &ops, out_file, wait_keyframe) < 0)
			break;
		}
		if (skip)
		    wait_for_start = 1;
		else
		    ops.pcr0 = ops.pcr1;
	    }
	}
	    break;
	case 2:
	    mpegts_cut_dvbapp(ts0, &ops, argv[1], wait_keyframe,
			      cps.first_pcr, cps.last_pcr,
			      out_file, out_s);
	    break;
	}
    }

    mpegts_delete_context(ts0);

    return 0;
}
