/*
 * $Id: $
 * Copyright 2011-2012 by Wolfram Gloger
 *
 *  This program is published under the GNU General Public license
 *
 */

#include <string.h>

#include "mpegts.h"
#include "mpegts-process.h"
#include "buffer.h"

#define STEP_PACKETS 2048

// callback for counting packets
static int
count_packet(void* arg0, unsigned char* pkt, int64_t pcr, int fl, long pkt_n)
{
    struct count_packet_s* cps = arg0;

    if (cps->first_pcr == -1)
	cps->first_pcr = pcr;
    if (pcr > cps->last_pcr)
	cps->last_pcr = pcr;
    ++cps->pid_histo[mpegts_get_pid(pkt)];
    if (fl & 1)
	++cps->n_keyframe;
    if (pkt_n*TS_PACKET_SIZE >= cps->nbytes)
	return -1;
    return 0;
}

// callback for obtaining pcr
struct obtain_pcr_s {
    int64_t pcr0;
    long pkt_n;
    int last;
};

static int
obtain_pcr(void* arg0, unsigned char* pkt, int64_t pcr, int fl, long pkt_n)
{
    struct obtain_pcr_s* ops = arg0;

    if (pcr != -1) {
	ops->pcr0 = pcr;
	ops->pkt_n = pkt_n;
	return ops->last ? 0 : -1;
    }
    return 0;
}

int64_t
mpegts_obtain_pcr(struct TSContext* ts, int last)
{
    struct obtain_pcr_s ops = { .pcr0 = -1, .last = last };
    int64_t o = mpegts_get_offset(ts);
    long np = buffer_process_packets(ts, obtain_pcr, &ops);

    if (np <= 0)
	return -1;
    mpegts_seek(ts, o + ops.pkt_n*TS_PACKET_SIZE);
    return (np <= 0) ? -1 : ops.pcr0;
}

int
mpegts_seek_pcr(struct TSContext* ts, int64_t pcr)
{
    int64_t pcr_min, off_min;
    int64_t pcr_max, off_max;
    int64_t pcr_est, off_est;

    fprintf(stderr, "* seeking to pcr=");
    print_pcr(stderr, pcr);
    fprintf(stderr, "\n");
    pcr_min = mpegts_obtain_pcr(ts, 0);
    if (pcr_min == -1) {
	mpegts_seek(ts, 0);
	pcr_min = mpegts_obtain_pcr(ts, 0);
    }
    off_min = mpegts_get_offset(ts);
    if (pcr_min < pcr) {
	int64_t step = 0;
	int last = 1;
	int fail;

	for (fail = 0;;) {
	    step -= STEP_PACKETS*TS_PACKET_SIZE;
	    if (mpegts_seek(ts, step) < 0)
		return -1;
	    pcr_max = mpegts_obtain_pcr(ts, last);
	    if (0) {
		fprintf(stderr, "pcr_max=");
		print_pcr(stderr, pcr_max);
		fprintf(stderr, "\n");
	    }
	    if (pcr_max == -1) {
		if (++fail > 10)
		    return -1;
	    } else {
		if (pcr_max >= pcr_min)
		    break;
	    }
	    last = 0;
	}
	off_max = mpegts_get_offset(ts);
    } else {
	pcr_max = pcr_min;
	off_max = off_min;
	mpegts_seek(ts, 0);
	pcr_min = mpegts_obtain_pcr(ts, 0);
	if (pcr_min == -1)
	    return -2;
	off_min = mpegts_get_offset(ts);
	if (pcr < pcr_min)
	    return 0;
    }
    for (;;) {
	fprintf(stderr,  "off_min=%10"PRId64" pcr_min=", off_min);
	print_pcr(stderr, pcr_min);
	fprintf(stderr, " off_max=%10"PRId64" pcr_max=", off_max);
	print_pcr(stderr, pcr_max);
	fprintf(stderr, "\n");

	// Interpolation step
	if (off_max - off_min < 2*TS_PACKET_SIZE)
	    break;
	off_est = (off_max - off_min)/(2*TS_PACKET_SIZE);
	mpegts_seek(ts, off_min + off_est*TS_PACKET_SIZE);
	pcr_est = mpegts_obtain_pcr(ts, 0);
	off_est = mpegts_get_offset(ts);
	if (pcr_est <= pcr) {
	    pcr_min = pcr_est;
	    off_min = off_est;
	} else {
	    pcr_max = pcr_est;
	    off_max = off_est;
	}
    }
    fprintf(stderr, "seek_off=%9"PRId64"\n", mpegts_get_offset(ts));
    return 0;
}

// compute the best PAT sid
long
mpegts_get_info(struct TSContext* ts0,
		int pat_probe_size,
		struct count_packet_s* cps)
{
    int64_t o = mpegts_get_offset(ts0);
    long np_read;

    memset(cps, 0, sizeof *cps);
    cps->first_pcr = -1;
    cps->nbytes    = pat_probe_size;

    // parse a couple of seconds to get pid_histo
    np_read = buffer_process_packets(ts0, count_packet, cps);

    // seek to end
    cps->nbytes = STEP_PACKETS*TS_PACKET_SIZE;
    if (mpegts_seek(ts0, -cps->nbytes) >= 0) {
	long np_end = buffer_process_packets(ts0, count_packet, cps);

	if (np_end >= 0)
	    np_read += np_end;
    } else
	fprintf(stderr, "Can't seek to end\n");

    // seek back to start
    if (mpegts_seek(ts0, o) != 0) {
	return -1;
    }

    return np_read;
}
