/* WG Sep 2011 */
/* $Id: mpegts.c,v 1.12 2013/04/11 12:01:01 wg Exp $
 * This program is published under the GNU General Public license
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <utime.h>

#include "mpegts.h"
#include "crc.h"

#define TS_SYNC_CODE   0x47
#define PAT_TABLE_SIZE 128

/* pids */
#define PAT_PID   0x0000
#define SDT_PID   0x0011

/* table ids */
#define PAT_TID   0x00
#define PMT_TID   0x02
#define SDT_TID   0x42

#define AV_RB16(p) (((*(const unsigned char*)(p)) << 8) | *((const unsigned char*)(p)+1))
#define AV_RB32(p) (((*(const unsigned char*)(p)) << 24) | *((const unsigned char*)(p)+1) << 16 | ((*(const unsigned char*)(p)+2) << 8) | *((const unsigned char*)(p)+3))
#define AV_WB16(p, v) do { \
	* (unsigned char*)(p)      = ((v) >> 8);\
	*((unsigned char*)(p) + 1) =  (v);      \
    } while(0)

typedef struct TSSection {
    unsigned char* section_buf;
    unsigned char* section_start; // start of data after the header
    int section_index, section_buf_size;
    int section_h_size;

    // used when changing/fixing the section
    uint8_t *section_fixed;
    int      section_bytes_fixed;

    // header
    uint8_t tid;
    uint16_t id;
    uint8_t version;
    uint8_t sec_num;
    uint8_t last_sec_num;
} TSSection;

static void section_free(TSSection* s)
{
    free(s->section_fixed);
    free(s->section_buf);
}

static int
section_data_fix(TSSection* ps, uint8_t *buf, int buf_size, int start)
{
    if (start)
	ps->section_bytes_fixed = 0;

    // compare with the original PAT section
    if (ps->section_bytes_fixed + buf_size > ps->section_index ||
	memcmp(ps->section_buf + ps->section_bytes_fixed, buf, buf_size)) {
	fprintf(stderr, "Section has changed within the source file, leaving\n");
    } else
	memcpy(buf, ps->section_fixed + ps->section_bytes_fixed, buf_size);
    ps->section_bytes_fixed += buf_size;
    return 0;
}

/************************************************************************/

struct TSContext {
    uint64_t packet_offset; // offset to the last synced but not yet parsed packet
    uint64_t keyframe_offset; // offset to last detected keyframe
    TSSection pat_sec, pmt_sec;
    // pat table
    int pat_sid[PAT_TABLE_SIZE], pat_pmt_pid[PAT_TABLE_SIZE];
    int pat_best_sid, pat_best_pmt;
    int pat_n;

    // pmt table
    int pcr_pid;
    int pmt_pid[PAT_TABLE_SIZE], codec_type[PAT_TABLE_SIZE];
    enum CodecId codec_id[8192];

    // parser
    uint64_t sps_packet_offset;
    uint32_t pstate;
    int x264_state;

    int fd;
    unsigned n_resync;
};

static void pat_data(struct TSContext* ts);
static void pmt_data(struct TSContext* ts);

static inline int get8(const uint8_t **pp, const uint8_t *p_end)
{
    const uint8_t *p;
    int c;

    p = *pp;
    if (p >= p_end)
        return -1;
    c = *p++;
    *pp = p;
    return c;
}

static inline int get16(const uint8_t **pp, const uint8_t *p_end)
{
    const uint8_t *p;
    int c;

    p = *pp;
    if ((p + 1) >= p_end)
        return -1;
    c = AV_RB16(p);
    p += 2;
    *pp = p;
    return c;
}

#if 0
/* read and allocate a DVB string preceeded by its length */
static char *getstr8(const uint8_t **pp, const uint8_t *p_end)
{
    int len;
    const uint8_t *p;
    char *str;

    p = *pp;
    len = get8(&p, p_end);
    if (len < 0)
        return NULL;
    if ((p + len) > p_end)
        return NULL;
    str = malloc(len + 1);
    if (!str)
        return NULL;
    memcpy(str, p, len);
    str[len] = '\0';
    p += len;
    *pp = p;
    return str;
}
#endif

static int parse_section_header(TSSection *h)
{
    const uint8_t *p = h->section_buf;
    const uint8_t *p_end = p + (h->section_h_size-4);
    int val;

    h->section_start = NULL;
    val = get8(&p, p_end);
    if (val < 0)
        return -1;
    assert(h->tid == val);
    p += 2;
    val = get16(&p, p_end);
    if (val < 0)
        return -1;
    h->id = val;
    val = get8(&p, p_end);
    if (val < 0)
        return -1;
    h->version = (val >> 1) & 0x1f;
    val = get8(&p, p_end);
    if (val < 0)
        return -1;
    h->sec_num = val;
    val = get8(&p, p_end);
    if (val < 0)
        return -1;
    h->last_sec_num = val;
    h->section_start = (uint8_t *)p;
    return 0;
}

static int
write_section_data(struct TSContext* ts, TSSection *s,
		   uint8_t *buf, int buf_size, int is_start)
{
    //fprintf(stderr, "section sz=%d, start=%d\n", buf_size, is_start);
    if (s->section_fixed) {
	assert(s->tid == PAT_TID);
	return section_data_fix(s, buf, buf_size, is_start);
    }

    if (is_start) {
        s->section_index = 0;
        s->section_h_size = -1;
    }

    // check if section_buf_size is big enough
    if (s->section_index + buf_size > s->section_buf_size) {
	int newsize = s->section_index + buf_size + 256;
	uint8_t *section_buf = realloc(s->section_buf, newsize);
	if (!section_buf)
	    return -1;
	s->section_buf      = section_buf;
	s->section_buf_size = newsize;
    }

    memcpy(s->section_buf + s->section_index, buf, buf_size);
    s->section_index += buf_size;

    /* compute section length if possible */
    if (s->section_h_size == -1 && s->section_index >= 3) {
	s->tid = *s->section_buf;
        s->section_h_size = (AV_RB16(s->section_buf + 1) & 0xfff) + 3;
    }

    if (s->section_h_size != -1 && s->section_index >= s->section_h_size) {
	if (get_crc32(-1, s->section_buf, s->section_h_size) != 0)
	    fprintf(stderr, "Invalid section crc\n");
	else {
	    assert(s->section_h_size >= 4);
	    //fprintf(stderr, "section sec_len=%zd\n", s->section_h_size);
	    if (parse_section_header(s) >= 0) {
		switch (s->tid) {
		case PAT_TID:
		    pat_data(ts);
		    break;
		case PMT_TID:
		    pmt_data(ts);
		    break;
		default:
		    fprintf(stderr, "not handling tid %x\n", s->tid);
		}
	    } else
		fprintf(stderr, "error parsing sec header\n");
	}
	return 1;
    }
    return 0;
}

/*------------------------------------------------------------------------*/

/* Read at most buffer_packets packets into buffer, and sync to start
 * code.  Return number of packets in the buffer.
 */
int
mpegts_read_packets(struct TSContext* ts,
		    unsigned char* buf, size_t buffer_packets)
{
    size_t rsize = TS_PACKET_SIZE*buffer_packets;
    size_t s = 0;
    size_t buffer_end;

    while (rsize > 0) {
	size_t r = read(ts->fd, buf+s, rsize);
	if (!r)
	    break;
	s += r;
	rsize -= r;
    }
    if (s < TS_PACKET_SIZE) {
	// discard partial packet
	buffer_end = 0;
	return 0;
    }

    buffer_end = s;

    // find sync code
    for (s=0; s<buffer_end-TS_PACKET_SIZE; ++s) {
	if (buf[s] != TS_SYNC_CODE) 
	    continue;
	if (s+TS_PACKET_SIZE < buffer_end &&
	    buf[s+TS_PACKET_SIZE] != TS_SYNC_CODE)
	    continue;
	if (s+2*TS_PACKET_SIZE < buffer_end &&
	    buf[s+2*TS_PACKET_SIZE] != TS_SYNC_CODE)
	    continue;
	if (s+3*TS_PACKET_SIZE < buffer_end &&
	    buf[s+3*TS_PACKET_SIZE] != TS_SYNC_CODE)
	    continue;
	break;
    }
    if (buf[s] != TS_SYNC_CODE) {
	ts->packet_offset += buffer_end;
	return 0;
    }
    if (s > 0) {
	++ts->n_resync;
	// discard unsynced bytes at the start
	fprintf(stderr, "* discarding %zd bytes\n", s);
	memmove(buf, buf+s, (buffer_end -= s));
	ts->packet_offset += s;
	// and try to fill up the final packet
	s = buffer_end % TS_PACKET_SIZE;
	if (s > 0) {
	    fprintf(stderr, "* appending %zd bytes\n", TS_PACKET_SIZE-s);
	    buffer_end += read(ts->fd, buf+buffer_end, TS_PACKET_SIZE-s);
	}
    }
    // count packets
    s = buffer_end/TS_PACKET_SIZE;
    while (s>0 && buf[(s-1)*TS_PACKET_SIZE] != TS_SYNC_CODE)
	--s;
    return s;
}

int
mpegts_get_pid(const uint8_t *pkt)
{
    return AV_RB16(pkt+1) & 0x1fff;
}

int64_t
get_pcr(const uint8_t* p)
{
    if ((p[3] & 0x20) && // adaptation field
	(p[4] >= 7) &&
	(p[5] & 0x10)) {
	int64_t pcr =
	    ((int64_t)p[6] << 25) |
	    ((int64_t)p[7] << 17) |
	    ((int64_t)p[8] << 9 ) |
	    ((int64_t)p[9] << 1 ) |
	    ((int64_t)p[10]>> 7 );
	pcr *= 300; // 90kHz -> 27MHz
	pcr += ((int64_t)(p[10] & 0x01) << 8) | (int64_t)p[11];
	return pcr;
    }
    return -1;
}

void
print_pcr(FILE* f, int64_t pcr)
{
    int64_t secE6 = pcr/(3*9);

    if (pcr < 0)
	fprintf(f, "---INVALID---");
    else
	fprintf(f, "%6"PRId64".%06"PRId64, secE6/1000000, secE6%1000000);
}

int64_t
atopcr(const char* s)
{
    int64_t pcr_90k, pcr;
    long sec, usec = 0;
    char* p;

    sec = strtol(s, &p, 10);
    if (*p == '.' || *p == ',') {
	char ss[7];
	strncpy(ss, p+1, 6);
	for (p=ss; (p-ss)<=5 && ('0'<= *p && *p<='9'); ++p)
	    ;
	while ((p-ss) <= 5)
	    *p++ = '0';
	*p = '\0';
	usec = strtol(ss, NULL, 10);
    }
    //fprintf(stderr, "sec=%ld usec=%ld\n", sec, usec);

    pcr_90k = ((int64_t)sec * 90000) + usec*9/100;
    if (pcr_90k & (-1ULL) << 33) {
	fprintf(stderr, "* truncating PCR %ldsec to 33bits in 90kHz\n", sec);
	pcr_90k &= ~((-1ULL) << 33);
	pcr = 3*100*pcr_90k;
	print_pcr(stderr,pcr);fprintf(stderr, "\n");
    } else {
	// compute 27MHz PCR directly
	pcr = ((int64_t)sec * 1000000) + usec;
	pcr *= 3*9;
    }

    return pcr;
}

/*------------------------------------------------------------------------*/

static void
pat_data(struct TSContext* ts)
{
    TSSection* s = &ts->pat_sec;
    const uint8_t *p     = s->section_start;
    const uint8_t *p_end = s->section_buf + (s->section_h_size-4);
    int sid, pmt_pid, i;

    for(i=0; i<PAT_TABLE_SIZE; ++i) {
        sid = get16(&p, p_end);
        if (sid < 0)
            break;
        pmt_pid = get16(&p, p_end);
        if (pmt_pid < 0)
            break;

        //fprintf(stderr, "sid=0x%x pmt_pid=0x%x\n", sid, pmt_pid&0x1FFF);
	ts->pat_sid[i] = sid;
	ts->pat_pmt_pid[i] = pmt_pid;
    }
    ts->pat_n = i;
}

// swap the first non-0 sid with ts->pat_best_sid
static int
pat_fix(struct TSContext* ts)
{
    TSSection* s = &ts->pat_sec;
    size_t sec_len       = s->section_h_size;
    const uint8_t *p, *p_end;
    int sid, pmt_pid, i;
    int sid0 = 0, pmt_pid0;

    assert(s->section_index > 0);
    assert(s->section_index <= s->section_buf_size);
    free(s->section_fixed);
    s->section_fixed = malloc(s->section_index);
    if (!s->section_fixed) {
	fprintf(stderr, "out of mem for pat_fix\n");
	return -1;
    }
    memcpy(s->section_fixed, s->section_buf, s->section_index);
    s->section_bytes_fixed = 0;
    p     = s->section_fixed + (s->section_start - s->section_buf);
    p_end = s->section_fixed + (sec_len-4);

    for(i=0;;) {
        sid = get16(&p, p_end);
        if (sid < 0)
            break;
        pmt_pid = get16(&p, p_end);
        if (pmt_pid < 0)
            break;

        //fprintf(stderr, "sid=0x%x pmt_pid=0x%x\n", sid, pmt_pid&0x1FFF);

	if (sid == 0x0000) {
	    /* NIT info */
	    continue;
	}

	if (i == 0 && sid != ts->pat_best_sid) {
	    sid0     = sid;
	    pmt_pid0 = pmt_pid;
	    if (0)
		fprintf(stderr, "Swapping sid0=%d with best_sid=%d\n",
			sid0, ts->pat_best_sid);
	    AV_WB16(p-4, ts->pat_best_sid);
	    AV_WB16(p-2, ts->pat_best_pmt);
	} else if (sid0>0 && sid == ts->pat_best_sid) {
	    AV_WB16(p-4, sid0);
	    AV_WB16(p-2, pmt_pid0);
	}
	++i;
    }

    if (sid0) { // update crc
       unsigned crc = get_crc32(-1, s->section_fixed, sec_len-4);

       //fprintf(stderr, "*crc=%8x\n", crc);
       s->section_fixed[sec_len-4] = crc        & 0xff;
       s->section_fixed[sec_len-3] = (crc >> 8) & 0xff;
       s->section_fixed[sec_len-2] = (crc >>16) & 0xff;
       s->section_fixed[sec_len-1] = (crc >>24) & 0xff;
    }

    return sid0;
}

/*------------------------------------------------------------------------*/

static int
parse_mpeg2video(struct TSContext* ts, int af1,
		 const uint8_t* p0, const uint8_t* p_end)
{
#if 0
    // Random access indicator
    if (af1 >= 0 && (af1 & 0x40)) {
	ts->keyframe_offset = ts->packet_offset;
	return 1;
    }
    return 0;
#endif
    int ret = 0;
    const uint8_t *p;
    uint32_t pstate = ts->pstate;

    for (p = p0; p < p_end; ++p) {
	if (pstate == 0x00000100) { // PICTURE start code
	    if (p+1 < p_end && ((p[1] >> 3) & 7) == 1) {
		ts->keyframe_offset = ts->packet_offset;
		ret = 1;
	    }
	}
	pstate = (pstate << 8) | *p;
    }
    ts->pstate = pstate;
    return ret;
}

static int
parse_x264(struct TSContext* ts, int af1,
	   const uint8_t* p0, const uint8_t* p_end)
{
    int ret = 0;
    const uint8_t *p;
    uint32_t pstate = ts->pstate;

    //fprintf(stderr, "parse_x264 %zd\n", p_end-p);
    for (p = p0; p < p_end; ++p) {
	if (pstate == 0x00000001) {
	    switch (*p & 0x0f) {
	    case 0x1:
		//fprintf(stderr, "x264 I or P frame at %zd\n", p-p0);
		if (ts->x264_state == 2) {
		    ts->keyframe_offset = ts->sps_packet_offset;
		    ts->x264_state = 0;
		    ret = 1;
		}
		break;
	    case 0x5:
		//fprintf(stderr, "x264 IDR at %zd\n", p-p0);
		break;
	    case 0x7:
		//fprintf(stderr, "x264 SPS header at %"PRIu64" %zd\n", ts->packet_offset, p-p0);
		ts->sps_packet_offset = ts->packet_offset;
		ts->x264_state = 1;
		break;
	    case 0x8:
		//fprintf(stderr, "x264 PPS at %zd\n", p-p0);
		if (ts->x264_state == 1)
		    ts->x264_state = 2;
		break;
	    }
	}
	pstate = (pstate << 8) | *p;
    }
    ts->pstate = pstate;
    return ret;
}

typedef int (*mpegts_parser)(struct TSContext*, int,
			     const uint8_t*, const uint8_t*);

static const mpegts_parser mpegts_parsers[] = {
    [CODEC_ID_MPEG2VIDEO] = parse_mpeg2video,
    [CODEC_ID_H264]       = parse_x264
};

/************************************************************************/

typedef struct {
    uint32_t stream_type;
    enum AVMediaType codec_type;
    enum CodecId codec_id;
} StreamType;

static const StreamType ISO_types[] = {
    { 0x01, AVMEDIA_TYPE_VIDEO, CODEC_ID_MPEG2VIDEO },
    { 0x02, AVMEDIA_TYPE_VIDEO, CODEC_ID_MPEG2VIDEO },
    { 0x03, AVMEDIA_TYPE_AUDIO,        CODEC_ID_MP3 },
    { 0x04, AVMEDIA_TYPE_AUDIO,        CODEC_ID_MP3 },
    { 0x0f, AVMEDIA_TYPE_AUDIO,        CODEC_ID_AAC },
    { 0x10, AVMEDIA_TYPE_VIDEO,      CODEC_ID_MPEG4 },
    { 0x11, AVMEDIA_TYPE_AUDIO,   CODEC_ID_AAC_LATM }, /* LATM syntax */
    { 0x1b, AVMEDIA_TYPE_VIDEO,       CODEC_ID_H264 },
    { 0xd1, AVMEDIA_TYPE_VIDEO,      CODEC_ID_DIRAC },
    { 0xea, AVMEDIA_TYPE_VIDEO,        CODEC_ID_VC1 },
    { 0 },
};

static int
mpegts_find_stream_type(int *codec_type, enum CodecId *codec_id,
			uint32_t stream_type, const StreamType *types)
{
    for (; types->stream_type; types++) {
        if (stream_type == types->stream_type) {
            *codec_type = types->codec_type;
            *codec_id   = types->codec_id;
            return 0;
        }
    }
    return -1;
}

const char*
mpegts_codec_name(enum CodecId codec_id)
{
    static const char* codec_name[] = {
	[CODEC_ID_NONE] = "none",
	[CODEC_ID_MP3] = "mp3",
	[CODEC_ID_AAC] = "aac",
	[CODEC_ID_AAC_LATM] = "aac latm",
	[CODEC_ID_MPEG2VIDEO] = "mpeg2 video",
	[CODEC_ID_MPEG4] = "mpeg4",
	[CODEC_ID_H264] = "h264",
	[CODEC_ID_DIRAC] = "dirac",
	[CODEC_ID_VC1] = "vc1"
    };
    return codec_name[codec_id];
}

static void
pmt_data(struct TSContext* ts)
{
    TSSection* s = &ts->pmt_sec;
    size_t sec_len       = s->section_h_size;
    const uint8_t *p     = s->section_start;
    const uint8_t *p_end = s->section_buf + (sec_len-4);
    const uint8_t *desc_list_end;
    int program_info_length, pid, stream_type;
    int desc_list_len;
    //uint32_t prog_reg_desc = 0; /* registration descriptor */
    //uint8_t *mp4_dec_config_descr = NULL;
    //int mp4_dec_config_descr_len = 0;
    //int mp4_es_id = 0;
    int i;

    ts->pcr_pid = get16(&p, p_end) & 0x1fff;
    if (ts->pcr_pid < 0)
        return;

    //fprintf(stderr, "pcr_pid=0x%x\n", ts->pcr_pid);

    program_info_length = get16(&p, p_end) & 0xfff;
    if (program_info_length < 0)
        return;
    while(program_info_length >= 2) {
        uint8_t tag, len;
        tag = get8(&p, p_end);
        len = get8(&p, p_end);

        //av_dlog(ts->stream, "program tag: 0x%02x len=%d\n", tag, len);

        if(len > program_info_length - 2)
            //something else is broken, exit the program_descriptors_loop
            break;
        program_info_length -= len + 2;
        if (tag == 0x1d) { // IOD descriptor
            get8(&p, p_end); // scope
            get8(&p, p_end); // label
            len -= 2;
            //mp4_read_iods(ts->stream, p, len, &mp4_es_id,
	    //              &mp4_dec_config_descr, &mp4_dec_config_descr_len);
        } else if (tag == 0x05 && len >= 4) { // registration descriptor
            //prog_reg_desc = AV_RB32(p);
	    p += 4;
            len -= 4;
        }
        p += len;
    }
    p += program_info_length;
    if (p >= p_end)
        goto out;

    for(i=0; i<PAT_TABLE_SIZE; ++i) {
        stream_type = get8(&p, p_end);
        if (stream_type < 0)
            break;
        pid = get16(&p, p_end) & 0x1fff;
        if (pid < 0)
            break;

	ts->pmt_pid[i] = pid;
	mpegts_find_stream_type(&ts->codec_type[i], &ts->codec_id[pid],
				stream_type, ISO_types);

        desc_list_len = get16(&p, p_end) & 0xfff;
        if (desc_list_len < 0)
            break;
        desc_list_end = p + desc_list_len;
        if (desc_list_end > p_end)
            break;
#if 0
        for(;;) {
            if (ff_parse_mpeg2_descriptor(ts->stream, st, stream_type, &p, desc_list_end,
                mp4_dec_config_descr_len, mp4_es_id, pid, mp4_dec_config_descr) < 0)
                break;

            if (prog_reg_desc == AV_RL32("HDMV") && stream_type == 0x83 && pes->sub_st) {
                ff_program_add_stream_index(ts->stream, h->id, pes->sub_st->index);
                pes->sub_st->codec->codec_tag = st->codec->codec_tag;
            }
        }
#endif
        p = desc_list_end;
    }
    if (i < PAT_TABLE_SIZE)
	ts->pmt_pid[i] = 0;

 out:
    ;
    //av_free(mp4_dec_config_descr);
}

/************************************************************************/

struct TSContext*
mpegts_new_context(const char* f)
{
    int fd;
    struct TSContext* ts;

    if (!strcmp(f, "-")) {
	fd = 0;
    } else {
	fd = open(f, O_RDONLY);
	if (fd < 0) {
	    fprintf(stderr, "Can't open input %s\n", f);
	    return NULL;
	}
    }
    ts = calloc(sizeof(struct TSContext), 1);
    if (!ts) {
	if (fd > 0)
	    close(fd);
	return NULL;
    }
    ts->keyframe_offset = -1;
    ts->pstate = 0xdeadbeef;
    ts->fd = fd;
    if (ts->fd > 0)
	ts->packet_offset = lseek(fd, 0, SEEK_CUR);
    return ts;
}

void
mpegts_delete_context(struct TSContext* ts)
{
    section_free(&ts->pmt_sec);
    section_free(&ts->pat_sec);
    if (ts->fd > 0)
	close(ts->fd);
    free(ts);
}

int
mpegts_parse_packet(struct TSContext* ts, unsigned char* pkt, int64_t pcr)
{
    int is_start, pid, afc, len, i, ret=0;

    assert(pkt[0] == TS_SYNC_CODE);
    is_start = pkt[1] & 0x40;
    pid = mpegts_get_pid(pkt);
    //    if (pid_histo)
    //	++pid_histo[pid];
#if 0
    fprintf(stderr, "Parse: pid=0x%x ", pid);
    print_pcr(stderr, pcr);
    fprintf(stderr, "\n");
#endif

    /* continuity check (currently not used) */
    //cc = (packet[3] & 0xf);

    /* adaptation field */
    afc = (pkt[3] >> 4) & 3;
    if (afc != 0 /* reserved value */ &&
	afc != 2 /* adaptation field only */) {
	int af1 = -1;
	unsigned char* p = pkt+4, * p_end = pkt+TS_PACKET_SIZE;

	if (afc == 3) {
	    if (p[0] >= 1)
		af1 = p[1];
	    /* skip adapation field */
	    p += p[0] + 1;
	}

	if (p < p_end) {
	    TSSection *s = NULL;

	    // treat the PAT and PMT sections only
	    if (pid == PAT_PID)
		s = &ts->pat_sec;
	    else {
		for (i=0; i<PAT_TABLE_SIZE && ts->pat_pmt_pid[i]; ++i) {
		    if (pid == (ts->pat_pmt_pid[i] & 0x1fff)) {
			s = &ts->pmt_sec;
			break;
		    }
		}
	    }

	    if (s) {
		if (is_start) {
		    /* pointer field present */
		    len = *p++;
		    if (p + len <= p_end) {
			if (len /*&& cc_ok*/) {
			    /* write remaining section bytes */
			    write_section_data(ts, s, p, len, 0);
			}
			p += len;
			if (p < p_end) {
			    write_section_data(ts, s, p, p_end - p, 1);
			}
		    }
		} else
		    write_section_data(ts, s, p, p_end - p, 0);
	    } else {
		// parse payload
		mpegts_parser mps = mpegts_parsers[ts->codec_id[pid]];

		if (mps)
		    ret = mps(ts, af1, p, p_end);
	    }
	}
    }

    ts->packet_offset += TS_PACKET_SIZE;
    return ret;
}

void
mpegts_compute_best_sid(struct TSContext* ts, const long pid_histo[])
{
    int i, sid0;
    long max;

    fprintf(stderr, "PAT_sid-pmt_pid\n");
    ts->pat_best_sid = 0;
    for (max=0, i=0; i<ts->pat_n; ++i) {
	int pmt_pid = ts->pat_pmt_pid[i] & 0x1FFF;
	fprintf(stderr, "%7d    %4x\n",
		ts->pat_sid[i], pmt_pid);
	if (pid_histo[pmt_pid] > max) {
	    max = pid_histo[pmt_pid];
	    ts->pat_best_sid = ts->pat_sid[i];
	    ts->pat_best_pmt = ts->pat_pmt_pid[i];
	}
    }
    fprintf(stderr, "---------------\n");
    fprintf(stderr, "PAT section_len %5d\n", ts->pat_sec.section_index);
    fprintf(stderr, "Best PAT sid is %d\n",  ts->pat_best_sid);
    fprintf(stderr, "Best PAT pmt is %5x\n", ts->pat_best_pmt & 0x1FFF);
    sid0 = pat_fix(ts);
    if (sid0)
	fprintf(stderr, "Fixing sid0 %d\n", sid0);
}

void
mpegts_print_pmt(const struct TSContext* ts)
{
    int i;

    fprintf(stderr, "PMT pcr_pid=%4x\n", ts->pcr_pid);
    fprintf(stderr, "==================\n");
    for(i=0; i<PAT_TABLE_SIZE && ts->pmt_pid[i]; ++i) {
	fprintf(stderr, "%4x %d %s\n",
		ts->pmt_pid[i], ts->codec_type[i], mpegts_codec_name(ts->codec_id[ts->pmt_pid[i]]));
    }
    if (ts->n_resync)
	fprintf(stderr, "*Resyncs: %7u\n", ts->n_resync);
}

int64_t
mpegts_seek(struct TSContext* ts, int64_t offset)
{
    off_t o = lseek(ts->fd, offset, offset>=0 ? SEEK_SET : SEEK_END);

    if (o < 0)
	return -1;
    ts->packet_offset = o;
    return o;
}

int64_t
mpegts_get_offset(const struct TSContext* ts)
{
    return lseek(ts->fd, 0, SEEK_CUR);
}

int64_t
mpegts_get_packet_offset(const struct TSContext* ts)
{
    return ts->packet_offset;
}

int64_t
mpegts_get_keyframe_offset(const struct TSContext* ts)
{
    return ts->keyframe_offset;
}

int
mpegts_get_utime(const struct TSContext* ts, struct utimbuf* ut)
{
    struct stat source_stat;
    int stat_res = fstat(ts->fd, &source_stat);

    if (stat_res == 0) {
	ut->actime  = source_stat.st_atime;
	ut->modtime = source_stat.st_mtime;
    }
    return stat_res;
}
